
#include "bigint.h"
using namespace std;

namespace AlgorithMan
{

    bigint::bigint()
    {
    }
    
    bigint::bigint(const bigint& x)
        : bits(x.bits)
    {
    }

    bigint::bigint(unsigned int value)
        : bits(8*sizeof(unsigned int), false)
    {
        for(vector<bool>::iterator p = bits.begin(); p != bits.end() && value > 0; p++)
        {
            *p = value % 2;
            value >>= 1;
        }
        normalize();
    }
    
    bigint::bigint(const string& value)
    {
        operator=(parse(value));
    }
    
    bigint::~bigint()
    {
    }

    void bigint::normalize()
    {
        while(bits.size() > 0 && bits[bits.size()-1] == false)
            bits.pop_back();
    }
    
    

    // ################################################################################################

    
    
    bigint bigint::parse(const std::string& s)
    {
        bigint result(0);
        static bigint ten(10);
        for(string::const_iterator i = s.begin(); i != s.end(); i++)
        {
            if(*i < '0' || '9' < *i)
                throw "not a number";
            result = result * ten + bigint((unsigned int)(*i-'0'));
        }
        
        result.normalize();
        return result;
    }
    
    
    string bigint::to_string(int base) const
    {
        switch(base)
        {
            case 2:
            {
                string result( bits.size() > 0 ? bits.size() : 1, '0');
                for(int j = bits.size()-1; j >= 0; --j)
                    result[j] = '0'+bits[bits.size()-1-j];
                return result;
            }
            
            case 10:
            {
                vector<short> temp(1,0);
                vector<short> powerof2(1,1);
        
                for(vector<bool>::const_iterator i = bits.begin(); i != bits.end(); i++)
                {
                    // add i-th powerof2 to temp, if i-th bit is true
                    if(*i)
                    {
                        int tempit = 0;
                        bool carrybit = false;
                        for(vector<short>::iterator powof2 = powerof2.begin(); powof2 != powerof2.end(); powof2++)
                        {
                            if(tempit >= temp.size())
                                temp.push_back(0);
                    
                            bool nextcarry = temp[tempit] + *powof2 + carrybit >= 10;
                            temp[tempit] = (temp[tempit] + *powof2 + carrybit) % 10;
                            carrybit = nextcarry;
                    
                            tempit++;
                        }
                        if(carrybit)
                            temp.push_back(1);
                    }
            
                    // multiply powerof2 with 2
                    bool carrybit = false;
                    for(vector<short>::iterator powof2 = powerof2.begin(); powof2 != powerof2.end(); powof2++)
                    {
                        bool nextcarry = (*powof2 * 2) + carrybit >= 10;
                        *powof2 = ((*powof2 * 2) + carrybit) % 10;
                        carrybit = nextcarry;
                    }
                    if(carrybit)
                        powerof2.push_back(1);

                }

                // build the string from the temp vector
                string result( temp.size() > 0 ? temp.size() : 1, '0');
                for(int j = result.size()-1; j >= 0; --j)
                    result[j] = '0'+temp[temp.size()-1-j];
                return result;
            }
            
            case 16:
            {
                string result( 2*(bits.size()/8 + (bits.size()%8!=0)), '0' );
                int position = result.size()-1;
                for(int i = 0; i < bits.size(); i+= 8)
                {
                    int byte = 0;
                    int bit = 1;
                    
                    for(int j = 0; j <= 8 && i+j < bits.size(); j++)
                    {
                        if(bits[i+j])
                            byte |= bit;
                        bit <<= 1;
                    }
                    
                    const char* hex = "0123456789ABCDEF";
                    
                    result[position--] = hex[byte & 0x0F];
                    result[position--] = hex[(byte & 0xF0) >> 4];
                }
                return result;
            }
            
            default:
                throw "bigint::to_string() only supports base 2,10 and 16!";
        }
    }
    
    

    // ################################################################################################

    
    
    bigint bigint::operator=(const bigint& x)
    {
        bits = x.bits;
        return *this;
    }

    
    
    bigint bigint::operator+(const bigint& x) const
    {
        int min = bits.size() > x.bits.size() ? x.bits.size() : bits.size();
        
        bigint result;
        bool carry = false;
        
        for(int i = 0; i < min; i++)
        {
            result.bits.push_back( (bits[i] + x.bits[i] + carry) % 2 );
            carry = (bits[i] + x.bits[i] + carry) / 2;
        }
        
        for(int i = min; i < bits.size(); i++)
        {
            result.bits.push_back(bits[i] != carry); // xor
            carry = bits[i] && carry;
        }

        for(int i = min; i < x.bits.size(); i++)
        {
            result.bits.push_back(x.bits[i] != carry); // xor
            carry = x.bits[i] && carry;
        }
        
        if(carry)
            result.bits.push_back(1);
        
        result.normalize();
        return result;
    }
    
    
    
    bigint bigint::operator-(const bigint& x) const
    {
        if(x.bits.size() > bits.size())
            return bigint(0);
        
        bigint result;
        bool carry = false;
        
        for(int i = 0; i < x.bits.size(); i++)
        {
            result.bits.push_back( bits[i] != (x.bits[i] != carry) ); // xor xor
            carry =   (bits[i] && x.bits[i] && carry)
                  || (!bits[i] && (x.bits[i] || carry));
        }
        
        for(int i = x.bits.size(); i < bits.size(); i++)
        {
            result.bits.push_back(bits[i] != carry); // xor
            carry = !bits[i] && carry;
        }

        if(carry)
            return bigint(0);
        
        result.normalize();
        return result;
    }
    
    
    
    bigint bigint::operator*(const bigint& x) const
    {
        bigint result;
        bigint xshifted = x;
        
        for(int i = 0; i < bits.size(); i++)
        {
            if(bits[i])
                result += xshifted;
            xshifted.bits.insert(xshifted.bits.begin(), 0);
        }

        result.normalize();
        return result;
    }
    
    
    
    void bigint::divmod(const bigint& x, bigint& div, bigint& mod) const
    {
        mod = *this;
        bigint temp = x;
        div = 0;

        int diff = this->bits.size() - x.bits.size();
        temp <<= diff;
        
        for(;diff >= 0; diff--)
        {
            if(mod >= temp)
            {
                div.bits.insert(div.bits.begin(),1);
                mod -= temp;
            }
            else
            {
                div <<= 1; // div.bits.insert(div.bits.begin(),0);
            }
            temp >>= 1;
        }
    }
    
    
    
    bigint bigint::operator/(const bigint& x) const
    {
        bigint div, mod;
        divmod(x, div, mod);
        div.normalize();
        return div;
    }
    
    
    
    bigint bigint::operator%(const bigint& x) const
    {
        bigint div, mod;
        divmod(x, div, mod);
        mod.normalize();
        return mod;
    }
    
    
    
    bigint bigint::operator<<(const bigint& x) const
    {
        bigint result(*this);
        for(bigint i = x; i > 0; i -= 1)
            result.bits.insert(result.bits.begin(), 0);
        return result;
    }
    
    
    
    bigint bigint::operator>>(const bigint& x) const
    {
        bigint result(*this);
        for(bigint i = x; i > 0; i -= 1)
            result.bits.erase(result.bits.begin());
        return result;
    }    

    
    
    // ################################################################################################

    
    
    void bigint::operator+=(const bigint& x)
    {
        operator=(operator+(x));
    }
    
    void bigint::operator-=(const bigint& x)
    {
        operator=(operator-(x));
    }
    
    void bigint::operator*=(const bigint& x)
    {
        operator=(operator*(x));
    }
    
    void bigint::operator/=(const bigint& x)
    {
        operator=(operator/(x));
    }

    void bigint::operator%=(const bigint& x)
    {
        operator=(operator%(x));
    }

    void bigint::operator<<=(const bigint& x)
    {
        operator=(operator<<(x));
    }
    
    void bigint::operator>>=(const bigint& x)
    {
        operator=(operator>>(x));
    }    

    
    
    // ################################################################################################

    

    bool bigint::operator==(const bigint& x) const
    {
        if(bits.size() != x.bits.size())
            return false;
        for(vector<bool>::const_iterator i = bits.begin(), j = x.bits.begin(); i != bits.end(); i++, j++)
            if(*i != *j)
                return false;
        return true;
    }
    
    bool bigint::operator!=(const bigint& x) const
    {
        return !operator==(x);
    }
    
    bool bigint::operator<=(const bigint& x) const
    {
        return !x.operator<(*this);
    }
    
    bool bigint::operator<(const bigint& x) const
    {
        if(bits.size() < x.bits.size())
            return true;
        if(bits.size() > x.bits.size())
            return false;        

        bool foo = false;
        for(vector<bool>::const_iterator i = bits.begin(), j = x.bits.begin(); i != bits.end(); i++, j++)
            if(*i != *j)
                foo = *j;
        return foo;
    }
    
    bool bigint::operator>(const bigint& x) const
    {
        return x.operator<(*this);
    }
    
    bool bigint::operator>=(const bigint& x) const
    {
        return x.operator<=(*this);
    }
    

    
    // ################################################################################################

    
    
    ostream& operator<<(ostream& os, const bigint& x)
    {
        return os << x.to_string();
    }
    
    istream& operator>>(istream& is, bigint& x)
    {
        string temp;
        is >> temp;
        x = bigint::parse(temp);
        return is;
    }    
    
}
